from datalogd import DataFilter, listify

class CSVDataFilter(DataFilter):
    """
    Format received data into a table of comma separated values.

    The column headers can be formatted using values from the data. For example, for the data::

        [
            {'type': 'temperature', 'id': '0', 'value': 22.35},
            {'type': 'humidity', 'id': '0', 'value': 55.0},
            {'type': 'temperature', 'id': '1', 'value': 25.80},
        ]

    and a node initialised using::

        sink [class=CSVDataSink keys="['value']", labels="['{type}_{id}'"];

    the output will be::

        temperature_0,humidity_0,temperature_1
        22.35,55.0,25.80

    Setting ``labels`` to ``None`` will disable the column headers.

    By default, the column headers will be generated on every receipt of data. To instead output the
    column headers only once on the first receipt of data, use the parameter ``header="once"``.
    Setting ``header=None`` will also disable the headers completely.

    :param keys: Name of data keys to format into columns of the CSV.
    :param labels: Labels for the column headers, which may contain mappings to data values.
    :param header: Display the column header ``"once"`` or ``"every"`` or ``None``.
    """

    def __init__(self, sinks=[], keys=["timestamp", "value"], labels=["timestamp", "{type}_{id}"], header="every"):
        super().__init__(sinks=sinks)
        self.keys = keys
        self.labels = labels
        self._generate_header = header

    def receive(self, data):
        """
        Accept ``data`` and format into a table of CSV.

        :param data: Data to format as CSV.
        """
        data = listify(data)
        # Generate column headers
        col_labels = []
        cols = []
        for d in data:
            if self.keys and isinstance(d, dict):
                # May be inefficient, but orders keys as requested
                for k_i, k in enumerate(self.keys):
                    for dk in d.keys():
                        if k == dk:
                            if self.labels:
                                col_labels.append(self.labels[k_i].format_map(d))
                            cols.append(listify(d[k]))
            else:
                if self.labels: col_labels.append("data")
                cols.append(d)
        if len(data) > 0:
            csv = ""
            if self._generate_header in ("once", "every"):
                # Generate column header
                if self.labels: csv += ",".join(col_labels) + "\n"
            if self._generate_header == "once":
                # Generate header once on first data receipt
                self._generate_header = None
            # Generate row from data
            for r_i in range(min(len(x) for x in cols)):
                csv += ",".join([str(col[r_i]) for col in cols]) + "\n"
            self.send(csv.rstrip("\n"))



import sys

from datalogd import DataSink

class PrintDataSink(DataSink):
    """
    Output data to standard-out or standard-error streams using the built-in
    python :meth:`print` method.

    :param end:    Line terminator.
    :param stream: Output stream to use, either "stdout" or "stderr".
    """
    def __init__(self, end="\n", stream="stdout"):
        self.end = end
        if stream == sys.stdout or stream == sys.stderr:
            self.stream = stream
        else:
            self.stream = sys.stderr if stream == "stderr" else sys.stdout

    def receive(self, data):
        """
        Accept ``data`` and print it out.

        :param data: Data to print.
        """
        print(data, end=self.end, file=self.stream)



from datalogd import DataFilter, listify

class AggregatorDataFilter(DataFilter):
    """
    Aggregates consecutively received data values into a list and passes the new
    array(s) on to sinks.

    The aggregated data can be sent at different intervals to that which it is
    received by using the ``send_every`` parameter. A value of 1 will send the
    aggregated data every time new data is received. A value of ``send_every``
    equal to ``buffer_size`` will result in the data being passed on only once
    the buffer is filled. A typical usage example would be for data which is
    received once every second, using ``buffer_size=86400`` and
    ``send_every=60`` to store up to 24 hours of data, and update sinks every
    minute.

    :param buffer_size: Maximum length for lists of aggregated data.
    :param send_every: Send data to connected sinks every *n* updates.
    :param aggregate: List of data keys for which the values should be aggregated.
    """
    def __init__(self, sinks=[], buffer_size=100, send_every=100, aggregate=["timestamp", "value"]):
        super().__init__(sinks=sinks)
        self.buffer_size = buffer_size
        self.aggregate = aggregate
        self.send_every = send_every if send_every > 1 else 1
        self._send_count = 0
        self._prev_data = []

    def receive(self, data):
        """
        Accept ``data``, aggregate selected values, and pass on aggregated data
        to any connected sinks.

        :param data: Data containing values to aggregate.
        """
        data = listify(data)
        for d_i, d in enumerate(data):
            # Check the last entry in the buffer. If it has the same keys/values
            # (except for the "value" entry) then append value to previous one.
            try:
                if self._prev_data[d_i].keys() == d.keys():
                    # Keys matching
                    for k, v in d.items():
                        if k not in self.aggregate and self._prev_data[d_i][k] != v:
                            # A (non-aggregating) value doesn't match previous entry
                            raise ValueError(f"Value of field {k} doesn't match previous entry.")
                    # All (non-aggregating) values match previous entry
                    for k, v in d.items():
                        if k in self.aggregate:
                            # An aggregating field, append to last entry
                            self._prev_data[d_i][k] = listify(self._prev_data[d_i][k])
                            self._prev_data[d_i][k].append(v)
                            # Simple buffer size control, probably horribly inefficient compared to a circular buffer
                            if len(self._prev_data[d_i][k]) > self.buffer_size: self._prev_data[d_i][k].pop(0)
                    # All aggregating fields now appended to last entry
                    continue
                else:
                    print("keys don't match")
            except (IndexError, ValueError) as ex:
                pass
            # Either first entry, or keys or (non-aggregating) values don't match
            # Convert the aggregating values into lists
            for k, v in d.items():
                if k in self.aggregate:
                    d[k] = listify(v)
            self._prev_data.append(d)
        self._send_count += 1
        if self._send_count >= self.send_every:
            self._send_count = 0
            self.send(self._prev_data)



from datalogd import DataFilter, listify

class CoolingPowerDataFilter(DataFilter):
    """
    Calculate power absorbed by cooling liquid through a system given flow rate and input and output
    temperatures.

    The calculation requires each receipt of data to contain two temperature entries and one flow
    rate entry. For example:

    .. code-block:: none

        [
            {"type": "temperature", "id": "A_0", "value": 12.34, "units": "C"},
            {"type": "temperature", "id": "A_1", "value": 23.45, "units": "C"},
            {"type": "flow_rate", "id": "A_0", "value": 0.456, "units": "L/min"}
        ]
    
    The IDs used to select the appropriate temperatures and flow rate may be given in the
    initialisation parameters. Temperatures should be in celsius (or kelvin), and flow rate should
    be in litres per minute. By default, the heat capacity and density of water will be used for
    calculations, but alternate values may supplied as a parameters. Heat capacity should be in
    J/kg/K, and density in g/mL.

    If all required data entries are present, a new entry with a ``coolingpower`` type will be
    added, and the data will then look like:

    .. code-block:: none

        [
            {"type": "temperature", "id": "A_0", "value": 12.34, "units": "C"},
            {"type": "temperature", "id": "A_1", "value": 23.45, "units": "C"},
            {"type": "flow_rate", "id": "A_0", "value": 0.456, "units": "L/min"},
            {"type": "coolingpower", "id": "A_0", "value": 353.28, "units": "W"}
        ]
    
    The ID used for the ``coolingpower`` entry may also be specified as an initialisation parameter.
    If the flow rate entry includes a ``"timestamp"`` field, its value will be copied to the
    cooling power entry.

    :param temperature_id_in: ID of temperature data for inlet liquid.
    :param temperature_id_out: ID of temperature data for outlet liquid.
    :param flow_rate_id: ID of liquid flow rate data.
    :param heatcapacity: Heat capacity of cooling liquid, in J/kg/K.
    :param density: Density of cooling liquid, in g/mL.
    :param coolingpower_id: ID to use for the calculated cooling power data.
    """
    def __init__(self, sinks=[], temperature_id_in="A_0", temperature_id_out="A_1", flow_rate_id="A_0", heatcapacity=4184, density=1.0, coolingpower_id="A_0"):
        super().__init__(sinks=[]) 
        self._t_id_in = temperature_id_in
        self._t_id_out = temperature_id_out
        self._f_id = flow_rate_id
        self._heatcapacity = heatcapacity
        self._density = density
        self._id = coolingpower_id

    def receive(self, data):
        """
        Accept the provided ``data`` and add a ``coolingpower`` entry calculated from flow rate and
        input and output temperatures.

        :param data: Data to calculate cooling power from.
        """
        data = listify(data)
        # Find appropriate t_in, t_out, flow rate among the data
        t_in = t_out = flow_rate = None
        for d in data:
            if isinstance(d, dict) and "type" in d and "id" in d and "value" in d:
                if d["type"] == "temperature":
                    if d["id"] == self._t_id_in:
                        t_in = d["value"]
                    elif d["id"] == self._t_id_out:
                        t_out = d["value"]
                elif d["type"] == "flow_rate":
                    if d["id"] == self._f_id:
                        flow_rate = d["value"]
                        # Attempt to copy the timestamp from flow rate if present
                        timestamp = d["timestamp"] if "timestamp" in d else None
            # Stop looking if we have found needed values already
            if not (t_in is None or t_out is None or flow_rate is None):
                break
        # Calculate cooling power if we have all required values
        if not (t_in is None or t_out is None or flow_rate is None):
            # Assumes density of liquid constant with temperature, probably ok up to about 50℃
            new_data = {
                "type": "coolingpower",
                "id": self._id,
                "value": round(self._heatcapacity*(self._density*flow_rate/60)*(t_out - t_in), 2),
                "units": "W"
            }
            # Add timestamp if we found one attached to the flow rate
            if not timestamp is None:
                new_data["timestamp"] = timestamp
            data.append(new_data)
        self.send(data)



import asyncio
import logging
from enum import Enum

from datalogd import DataSource

try:
    import visa
    from ThorlabsPM100 import ThorlabsPM100
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("ThorlabsPM100 and/or visa module not found. Install it with \"pip install pyvisa pyvisa-py pyusb ThorlabsPM100\" or similar. A VISA backend must also be present, use pyvisa-py if the native NI libraries are not installed.")
else:
    # Required modules present, continue loading rest of this module

    class ThorlabsPM100DataSource(DataSource):
        """
        Provide data from a Thorlabs PM100 laser power meter.

        This uses the VISA protocol over USB. On Linux, read/write permissions
        to the power meter device must be granted. This can be done with a udev
        rule such as:

        .. code-block:: none
            :caption: ``/etc/udev/rules.d/52-thorlabspm100.rules``

            SUBSYSTEMS=="usb", ACTION=="add", ATTRS{idVendor}=="1313", ATTRS{idProduct}=="8078", OWNER="root", GROUP="usbusers", MODE="0666"

        where the ``idVendor`` and ``idProduct`` fields should match that listed
        from running ``lsusb``. The ``usbusers`` group must be created and the
        user added to it:

        .. code-block:: bash

            groupadd usbusers
            usermod -aG usbusers yourusername

        A reboot will then ensure permissions are set and the user is a part of
        the group.

        :param serialnumber: Serial number of power meter to use. If ``None``, will use the first device found.
        :param interval: How often to poll the sensors, in seconds.
        """
        def __init__(self, sinks=[], serialnumber=None, usb_vid="0x1313", usb_pid="0x8078", interval=1.0):
            super().__init__(sinks=sinks)
            self.log = logging.getLogger("ThorlabsPM100DataSource")
            self.interval = interval
            self.rm = visa.ResourceManager()
            if self.rm.visalib.library_path == "unset":
                # Native python VISA library, USB VID and PID in decimal, has extra field
                # Here, 4883 == vendorID == 0x1313, 32888 == productID == 0x8078
                try:
                    usb_vid = int(usb_vid, 16)
                    usb_pid = int(usb_pid, 16)
                except Exception as ex:
                    pass
                res = self.rm.list_resources("USB0::{}::{}::{}::0::INSTR".format(usb_vid, usb_pid, serialnumber if serialnumber else "?*"))
            else:
                # NI VISA library (probably) in use, USB VID and PID are in hex, also extra field missing
                res = self.rm.list_resources("USB0::{}::{}::{}::INSTR".format(usb_vid, usb_pid, serialnumber if serialnumber else "?*"))
            if len(res) == 0:
                self.log.warning("Could not find a Thorlabs PM100 device{}.".format(f" with serial {serialnumber}" if serialnumber else ""))
                self.inst = None
                self.pm100 = None
            else:
                try:
                    self.inst = self.rm.open_resource(res[0])
                    self.pm100 = ThorlabsPM100(self.inst)
                    self.log.info(f"Initialised Thorlabs PM100 device: {self.inst.resource_info[0].resource_name}.")
                    # Queue first call of update routine
                    asyncio.get_event_loop().call_soon(self.read_power)
                except Exception as ex:
                    self.log.warning("Could not initialise Thorlabs PM100 device: {}".format(ex))

        def close(self):
            """
            Close the connection to the power meter.
            """
            self.rm.close()

        def read_power(self):
            """
            Read power and send data to any connected sinks.
            """
            loop = asyncio.get_event_loop()
            try:
                data = {"type": "power", "source": "PM100", "id": self.inst.resource_info[0].resource_name, "value": self.pm100.read}
                self.send(data)
            except Exception as ex:
                self.log.warning("Could not read power from Thorlabs PM100 device.")
            # Reschedule next update
            loop.call_later(self.interval, self.read_power)



import logging
from datetime import datetime

from datalogd import DataSink, listify

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("matplotlib module not found. Install it with \"pip install matplotlib\" or similar.")
else:
    # Required modules present, continue loading rest of this module

    class MatplotlibDataSink(DataSink):
        """
        .. note:: This plugin is still a work in progress, and is really only at
            the proof-of-concept stage.
        """

        def __init__(self, filename="plot.pdf", keys=["timestamp", "value"], labels=["timestamp", "{type}_{id}"]):
            self.filename = filename
            self.keys = keys
            self.labels = labels
            # Make the figure canvas
            self.fig, self.ax = plt.subplots(1, 1, figsize=(4.5, 3.5))

        def receive(self, data):
            # TODO: Restrict to pairs of columns of data?
            data = listify(data)
            col_labels = []
            cols = []
            for d in data:
                if self.keys and isinstance(d, dict):
                    # May be inefficient, but orders keys as requested
                    for k_i, k in enumerate(self.keys):
                        for dk in d.keys():
                            if k == dk:
                                col_labels.append(self.labels[k_i].format_map(d))
                                cols.append(listify(d[k]))
                else:
                    col_labels.append("data")
                    cols.append(d)

            # Remove old traces
            self.ax.clear()
            # Not guaranteed pairs columns will have equal lengths...
            for col_i, col in enumerate(cols[::2]):
                row_count = min(len(cols[2*col_i]), len(cols[2*col_i+1]))
                self.ax.plot(cols[2*col_i][:row_count], cols[2*col_i+1][:row_count], label=col_labels[2*col_i+1])

            if type(cols[0][0]) == datetime:
                self.fig.autofmt_xdate()

            self.ax.legend(frameon=False)
            self.fig.tight_layout()
            self.fig.savefig(self.filename)



import logging

from datalogd import DataSink, listify

class LoggingDataSink(DataSink):
    """
    Output data using python's :mod:`logging` system.

    Each item of data is output in a separate line, and the formatting can be
    controlled using the ``header`` and ``indent`` parameters.

    :param level: The `logging level <https://docs.python.org/3/library/logging.html#logging-levels>`_ to use for the output.
    :param header: Line of header text preceeding the logged data.
    :param indent: Prefix applied to each line of logged data.
    """
    def __init__(self, level=logging.INFO, header="Data received:", indent="  "):
        self.log = logging.getLogger("LoggingDataSink")
        self.level = level
        self.header = header
        self.indent = indent

    def receive(self, data):
        """
        Accept the provided data and output it using python's :mod:`logging`
        system.

        :param data: Data to log.
        """
        if data is None or data == []: return
        data = listify(data)
        if self.header: self.log.log(self.level, self.header)
        for d in data:
            self.log.log(self.level, f"{self.indent}{d}")



import asyncio
import logging
from ctypes import c_int8, c_int16, c_int32, c_float, byref
from datetime import datetime, timezone, timedelta

from datalogd import DataSource

try:
    from picosdk.usbtc08 import usbtc08 as tc08
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("picosdk module not found. Install it with \"pip install picosdk\" or similar.")
else:
    class PicoTC08DataSource(DataSource):
        """
        Obtain readings from a Pico Technologies TC-08 USB data logging device.

        The drivers and libraries (such as libusbtc08.so on Linux, usbtc08.dll on Windows) from PicoTech
        must be installed into a system library directory, and the ``picosdk`` python wrappers package
        must be on the system (with ``pip install picosdk`` or similar).

        The ``interval`` parameter determines how often data will be obtained from the sensors, in
        seconds. The minimum interval time is about 0.2 s for a single probe and 0.9 s for all eight.

        The ``mains_rejection`` parameter filters out either 50 Hz or 60 Hz interference from mains
        power. The frequency is selected as either ``"50Hz"`` or ``"60Hz"``.

        The ``probes`` parameter is a list of probe to initialise. Each element is itself a list of
        the form ``[number, label, type, units]``, where probe numbers are unique integers from 1 to
        8 corresponding to an input channel on the device. Probe labels can be any valid string.
        Valid probe thermocouple types are ``"B"``, ``"E"``, ``"J"``, ``"K"``, ``"N"``, ``"R"``,
        ``"S"``, ``"T"``, or ``"X"``, where ``"X""`` indicates a raw voltage reading. Units are one
        of Celsius, Fahrenheit, Kelvin, Rankine specified as ``"C"``, ``"F"``, ``"K"`` or ``"R"``.
        For the ``"X"`` probe type, readings will always be returned in millivolts.
        """
        def __init__(self, sinks=[], interval=1.0, mains_rejection="50Hz", probes=[
                [1, "Channel_1", "K", "C"],
                [2, "Channel_2", "K", "C"],
                [3, "Channel_3", "K", "C"],
                [4, "Channel_4", "K", "C"],
                [5, "Channel_5", "K", "C"],
                [6, "Channel_6", "K", "C"],
                [7, "Channel_7", "K", "C"],
                [8, "Channel_8", "K", "C"],
            ]):
            super().__init__(sinks=sinks)
            self._log = logging.getLogger("PicoTC08")
            self._log.info("Initialising Pico TC-08 device.")
            self._handle = tc08.usb_tc08_open_unit()
            if self._handle <= 0:
                raise RuntimeError("Unable to initialise Pico TC-08 USB device.")
            # Set mains power frequency rejection
            tc08.usb_tc08_set_mains(self._handle, 1 if mains_rejection == "60Hz" else 0)
            # Configure each specified channel, keep configured probes in a dictionary by channel
            self._probes = {}
            for i, probe in enumerate(probes):
                try:
                    probe_n = int(probe[0])
                    probe_label = str(probe[1])
                    probe_type = str(probe[2])[0]
                    probe_units = str(probe[3])[0]
                    status = tc08.usb_tc08_set_channel(self._handle, probe_n, ord(probe_type))
                    if status == 1:
                        self._probes[probe_n] = {"type": probe_type, "label": probe_label, "units": probe_units}
                        self._log.info(f"Configured TC-08 channel {probe_n} as type {probe_type}.")
                    else:
                        self._log.error(f"Error configuring TC-08 channel {probe_n} as type {probe_type}.")
                except Exception as ex:
                    self._log.exception("Error configuring TC-08 channel using probe configuration \"{probe}\"")
            # Should check that at least one probe was configured...
            if len(self._probes) <= 0:
                tc08.usb_tc08_close_unit(self._handle)
                self._handle = 0
                raise RuntimeError("No TC-08 channels were successfully configured.")
            # Get minimum possible sampling interval, in ms
            self._interval = max(int(1000*interval), tc08.usb_tc08_get_minimum_interval_ms(self._handle))
            # Start the device
            interval = tc08.usb_tc08_run(self._handle, self._interval)
            if interval == 0:
                tc08.usb_tc08_close_unit(self._handle)
                self._handle = 0
                raise RuntimeError("Error starting Pico TC-08 device.")
            else:
                self._interval = interval/1000
            # Queue first call of update routine
            asyncio.get_event_loop().call_soon(self.acquire_data)
            # Record local time of acquisition start
            self._time_start = datetime.now(timezone.utc).astimezone()


        def acquire_data(self):
            """
            Acquire data from the TC-08 streaming buffer.
            """
            loop = asyncio.get_running_loop()
            data = []
            for probe_n, probe in self._probes.items():
                temps = (c_float * 128)()
                times = (c_int32 * 128)()
                overflow = c_int16()
                count = tc08.usb_tc08_get_temp(
                    self._handle,
                    byref(temps),
                    byref(times),
                    128,
                    byref(overflow),
                    probe_n,
                    "CFKR".index(probe["units"]) if probe["units"] in "CFKR" else 0,
                    1)
                if count < 0:
                    self._log.error("Error acquiring data from Pico TC-08.")
                    tc08.usb_tc08_close_unit(self._handle)
                    self._handle = 0
                    return
                # Loop through each record
                for i in range(count):
                    data.append({
                        "timestamp" : (self._time_start + timedelta(milliseconds=times[i])).isoformat(),
                        "type" : "analog" if probe["type"] == "X" else "temperature",
                        "source" : "TC-08",
                        "id" : probe_n,
                        "label" : probe["label"],
                        "probe" : probe["type"],
                        "units" : "mV" if probe["type"] == "X" else probe["units"],
                        "value": temps[i]})
            self.send(data)
            # Reschedule next update
            loop.call_later(self._interval, self.acquire_data)


        def close(self):
            """
            Close the connection to the Pico TC-08 device.
            """
            if self._handle:
                tc08.usb_tc08_close_unit(self._handle)
                self._handle = 0




import logging
from datetime import datetime, timezone

from datalogd import DataSink, listify

try:
    from influxdb_client import InfluxDBClient
    from influxdb_client.client.write_api import SYNCHRONOUS
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("influxdb_client module not found. Install it with \"pip install influxdb_client\" or similar.")
else:
    # Required modules present, continue loading rest of this module

    class InfluxDB2DataSink(DataSink):
        """
        Connection to a InfluxDB 2.x (or 1.8+) database for storing time-series data.

        Note that this doesn't actually run the InfluxDB database service, but simply connects to an
        existing InfluxDB database via a network (or localhost) connection. See the `getting started
        <https://docs.influxdata.com/influxdb/v2.4/get-started/>`__ documentation for details on
        configuring a new database server.

        The ``url`` parameter should be a string specifying the protocol, server ip or name, and
        port. For example, ``url="http://localhost:8086"``.

        The authentication ``token`` parameter needs to be specified to allow commits to the
        database. See the `token <https://docs.influxdata.com/influxdb/v2.4/security/tokens/>`__
        documentation to see how to create and obtain tokens.

        Parameters for ``org``, ``bucket`` must correspond to a valid organisation and bucket
        created in the database for which the authentication token is valid. See the documentation
        for `organisations <https://docs.influxdata.com/influxdb/v2.4/organizations/>`__ and
        `buckets <https://docs.influxdata.com/influxdb/v2.4/organizations/buckets/>`__ for details.

        The ``measurement`` parameter specifies the data point measurement (or "table") the data
        will be entered into, and does not need to already exist. See the documentation on `data
        elements
        <https://docs.influxdata.com/influxdb/v2.4/reference/key-concepts/data-elements/#measurement>`__
        for details.

        A ``run_id`` parameter may be passed which will be added as a tag to the data points. It may
        be used to identify data obtained from this particular run of the data logging session. If
        no value is provided, a value will be generated from a YYYYMMDD-HHMMSS formatted time stamp.

        The data point field key will be attempted to be determined automatically from the incoming
        data dictionaries. If the data dictionary contains a ``name`` or ``label`` key, then its
        value will be used as the database point field key. Alternatively, a field key will be
        generated from the values of ``type`` and ``id`` if present. Finally, a default field key of
        ``data`` will be used. To instead specify the data entry which should provide the field key,
        specify it as the ``field_key`` parameter. If the field is specified by a parameter or taken
        from a name or label, then those will not also be included in the entry's database keys.
        However, if the field name is automatically built from type and id values, these will still
        be part of the entries keys.

        Similarly, the data point field value will use the value from the incoming data dictionary's
        ``value`` field if present. To instead specify the data entry which should provide the field
        value, specify it as the ``field_value`` parameter. The value won't also appear in the
        database entry's keys.

        :param url: Protocol, host name or IP address, and port number of InfluxDB server.
        :param token: API token used to authenticate with the InfluxDB server.
        :param org: Name of InfluxDB organisation in which to store data.
        :param bucket: Name of InfluxDB bucket in which to store data.
        :param measurement: Name for the InfluxDB measurement session.
        :param run_id: A tag to identify commits from this run.
        :param field_key: A field from the incoming data used to determine the data point field key.
        :param field_value: A field from the incoming data used to determine the data point field value.
        """
        def __init__(self, url="http://localhost:8086", token="", org="default", bucket="default", measurement="measurement", run_id=None, field_key=None, field_value=None):
            self.log = logging.getLogger("InfluxDB2DataSink")
            try:
                self.client = InfluxDBClient(url, token=token, org=org)
            except Exception as ex:
                self.log.exception("Unable to make connection to InfluxDB database.")
            self.write_api = self.client.write_api(write_options=SYNCHRONOUS)
            self.org = org
            self.bucket = bucket
            self.measurement = measurement
            if run_id is None:
                self.run_id = datetime.now(timezone.utc).astimezone().strftime("%Y%m%d-%H%M%S")
            else:
                self.run_id = run_id
            self.field_key = field_key
            self.field_value = field_value

        def receive(self, data):
            """
            Commit data to the InfluxDB database.

            Multiple items of data can be submitted at once if ``data`` is a
            list. A typical format of ``data`` would be::

                [
                    {'type': 'temperature', 'id': '0', 'value': 22.35},
                    {'type': 'humidity', 'id': '0', 'value': 55.0},
                    {'type': 'temperature', 'id': '1', 'value': 25.80},
                ]

            In the above case (assuming the ``field_key`` and ``field_value`` parameters were not
            supplied when initialising the plugin), the InfluxDB data point field would be generated
            as ``<type>_<id> = <value>``, and only the global ``run_id`` parameter would be entered
            into the data point keys.

            If a ``name`` or ``label`` field is present, then it will instead be used as the
            InfluxDB data point field key. For example::

                [
                    {'name': 'Temperature', 'type': 'temperature', 'id': '0', 'value': 22.35},
                    {'name': 'Humidity', 'type': 'humidity', 'id': '0', 'value': 55.0},
                ]
            
            In this case, the InfluxDB data point field would be generated as ``<name> = <value>``,
            and the remaining fields (``type`` and ``id``) would be added as data point field keys,
            along with the ``run_id``.


            A timestamp for the commit will be generated using the current system clock if a
            "timestamp" field does not already exist.

            :param data: Data to commit to the database.
            """
            if data is None or data == []: return
            data = listify(data)

            # Loop through each element in data list
            for d in data:

                # Start building the structure to enter into database
                datapoint = {
                    "measurement": self.measurement,
                    "tags": {"run_id": self.run_id},
                    "fields": {},
                }

                # Attempt to determine a suitable field name
                if self.field_key is not None and self.field_key in d.keys():
                    # If database field key was specified, use it (don't also add as a tag)
                    k = str(d.pop(self.field_key))
                elif "name" in d.keys():
                    # Format as <name> = <value> (don't also add as a tag)
                    k = str(d.pop("name"))
                elif "label" in d.keys():
                    # Format as <label> = <value> (don't also add as a tag)
                    k = str(d.pop("label"))
                elif "type" in d.keys():
                    # Format as <type>_<id> = <value> (but keep type, id as tags)
                    k = str(d["type"]) + ("_" + str(d["id"])) if "id" in d.keys() else ""
                else:
                    k = "data"
                # Attempt to determine the correct field value
                if self.field_value is not None and self.field_value in d.keys():
                    # If database field value was specified, use it
                    v = str(d.pop(self.field_value))
                elif "value" in d.keys():
                    v = d.pop("value")
                else:
                    v = None
                # Add the field to the data point
                datapoint["fields"][k] = v

                # Use the provided timestamp if available, else create one
                if "timestamp" in d.keys():
                    timestamp = d.pop("timestamp")
                    if type(timestamp) == datetime:
                        datapoint["time"] = timestamp.isoformat()
                    else:
                        datapoint["time"] = timestamp
                else:
                    datapoint["time"] =  datetime.now(timezone.utc).astimezone().isoformat()
                
                # Add any entries which weren't used in the field to the tags
                datapoint["tags"].update(d)
                
                # Send data point out to database
                try:
                    self.log.debug(f"Committing: {datapoint}")
                    self.write_api.write(self.bucket, self.org, datapoint)
                except Exception as ex:
                    self.log.warning("Unable to commit data to InfluxDB database.", exc_info=True)

        def close(self):
            """
            Close the connection to the database.
            """
            self.client.close()



from  math import exp

from datalogd import DataFilter, listify

class FlowSensorCalibrationDataFilter(DataFilter):
    """
    Use a pulse counter's counts per second to compute a liquid flow rate in litres per minute using
    an experimentally determined calibration function.

    A flow sensor has a spinning turbine and outputs pulses due to the flow rate of the liquid.
    However, the pulse rate will not be directly proportional to the flow rate (each pulse does not
    correspond to a fixed volume of liquid). A calibration curve can be constructed by measuring the
    number of pulses emitted over time for a given volume of liquid at a range of different flow
    rates. A plot of counts per litre versus counts per minute displays the characteristics of the
    sensor. Fitting the points to a curve of the form :math:`f(x) = a(1-\exp(-k(x-x_0)^b))` will
    provide the required calibration parameters.
    
    The default parameters (``a=5975, k=0.173734, x0=0, b=0.284333``) convert from counts per second
    to litres per minute for the YF-S401 flow sensor, and may be compatible with models from the
    same family such as YF-S402 and YF-S402B. A similar, smaller sensor common in automatic coffee
    machines, model number FM-HL3012C, was found to have parameters of ``a=1950, k=0.0965882, x0=0,
    b=0.721649``.

    The original ``count_rate`` entry in the data will be preserved, with the calculated
    ``flow_rate`` being appended as a new data entry.

    :param counter_rate_id: ID field to match to the data.
    :param a: Parameter :math:`a` in calibration function.
    :param k: Parameter :math:`k` in calibration function.
    :param x0: Parameter :math:`x_0` in calibration function.
    :param b: Parameter :math:`b` in calibration function.
    :param units: New units for the data.
    """
    def __init__(self, sinks=[], counter_rate_id="A_0", a=5975, k=0.173734, x0=0, b=0.284333):
        super().__init__(sinks=[])
        self._counter_rate_id = counter_rate_id
        self._a = a
        self._k = k
        self._x0 = x0
        self._b = b


    def receive(self, data):
        """
        Accept the provided ``data`` and compute a flow rate using the calibration function.

        :param data: Data to calculate flow rate from.
        """
        data = listify(data)
        # Find appropriate count rate among the data
        for d in data:
            try:
                if d["type"] == "counter_rate" and d["id"] == self._counter_rate_id:
                    # Get count rate, convert to counts per minute
                    cpm = 60.0*d["value"]
                    d_new = d.copy()
                    d_new["type"] = "flow_rate"
                    d_new["units"] = "L/min"
                    # Apply calibration function (for non-zero flow rates)
                    if cpm <= self._x0:
                        d_new["value"] = 0.0
                    else:
                        d_new["value"] = cpm/(self._a*(1 - exp(-self._k*(cpm - self._x0)**self._b)))
                    data.append(d_new)
            except KeyError:
                # That didn't work, not the data we're looking for
                continue
        self.send(data)



import logging
import re
from datetime import datetime, timezone
from multiprocessing import Process, Queue
from queue import Empty

from datalogd import DataSink, listify

try:
    from PySide6 import QtCore
    from PySide6 import QtWidgets

    import pyqtgraph as pg
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("pyqtgraph module not found. Install it with \"pip install pyqtgraph pyside6\" or similar.")
else:
    # Required modules present, continue loading rest of this module
    import numpy as np


    class PyqtgraphDataSink(DataSink):
        """
        Plot data in realtime in a pyqtgraph window.

        Multiple plot areas may be defined which will be stacked in rows with (by default) linked
        time axes. Each plot area may itself have multiple traces contained within. The complete
        plot configuration is defined in the initialisation parameters. The data to use for each
        trace is selected by matching a series of key-value pairs, in a similar manner to the
        :data:`~datalogd.plugins.keyval_datafilter.KeyValDataFilter`.

        A limited number of data points are stored to be plotted, after which the oldest data points
        will be discarded to make way for incoming data. The number of data points can be specified
        with the ``npoints`` parameter, with a default of 2048.

        The plot layout is described by the ``plotlayout`` parameter. As python code:

        .. code-block:: python

            plotlayout = [
                # List of plot panels
                {
                    # Plot 1 panel definition
                    'ylabel': 'Value (a.u)',
                    'traces': [
                        # List of trace definitions for this plot panel
                        {
                            # Trace 1 definition
                            'name': 'Trace 1',
                            'pen': [255, 255, 0],
                            'selector': [
                                # list of key-value pairs to match to data (same as KeyValDataFilter)
                                ['type', 'analog'],
                                ['id', '.*0']
                            ]
                        }, # ... possibly more trace definitions
                    ]
                }, # ... possibly more plot definitions
            ]
        
        In the connection graph configuration the ``plotlayout`` data structure must be a string
        formatted as JSON.

        Note that any fields present in a trace definition (such as ``'name'`` and ``'pen'``) are
        passed to the pyqtgraph :data:`~pyqtgraph.graphicsItems.PlotDataItem` initialisation which
        may be used to customise the trace, such as defining line color or changing to a scatter
        plot.

        Passing the parameter ``xlink=False`` will unlink the time axes of the plots, so changes to
        the view of one plot will not affect the others.

        By default, a crosshair will be shown under the mouse pointer. Values for each trace at the
        crosshair x position are shown in the legend, and the y position of the crosshair will be
        shown to the right of the plot. To disable this functionality, pass ``crosshair=False``.

        Any additional parameters are passed to the pyqtgraph
        :data:`~pyqtgraph.widgets.GraphicsLayoutWidget` initialisation, which can be used to
        customise the plot window. For example, changing the window title and size with
        ``title="Plots"`` and ``size=[1000, 600]``.

        :param npoints: Maximum number of data points for a trace.
        :param title: String for title of the plot window.
        :param size: Tuple of (height, width) of the plot window.
        :param plotlayout: Data structure describing the plot layout and traces.
        :param xlink: Boolean, link the time axes of the plots.
        :param crosshair: Boolean, show the crosshair under the mouse pointer.
        """
        def __init__(self, **kwargs):
            
            self.q = Queue()
            kwargs.update({"data_queue": self.q})
            self.appprocess = Process(target=self._exec_qt_app, kwargs=kwargs, name="PlotWindow")
            self.appprocess.start()

        def close(self):
            """
            Signal the pyqtgraph application to close when the application is shutting down.
            """
            #print("Waiting for QApplication to close...")
            self.appprocess.terminate()
            self.appprocess.join()
            #print("QApplication closed.")

        def receive(self, data):
            """
            Accept the provided ``data`` and pass it to the pyqtgraph :data:`PlotWindow` for
            display.
            """
            if self.appprocess.is_alive():
                data = listify(data)
                self.q.put(data)

        def _exec_qt_app(self, **kwargs):
            self.plotwindow = PlotWindow(**kwargs, show=True)
            app = QtWidgets.QApplication.instance()
            app.exec()


    class PlotWindow(pg.GraphicsLayoutWidget):

        def __init__(self, parent=None, data_queue=None, npoints=2**11, plotlayout=None, xlink=True, crosshair=True, **kwargs):
            super().__init__(**kwargs)
            
            # Queue object for receiving data through
            self.data_queue = data_queue

            # Default plot layout if none specified
            if plotlayout is None: plotlayout = [{}]
            if not type(plotlayout) == list:
                raise RuntimeError(f"plotlayout should be a list of dicts, not \"{type(plotlayout).__name__}\"")

            pg.setConfigOption("antialias", True)

            # Parse plot layout data structure to construct plots
            # The data structure looks like:
            # self.plots = [
            #   (PlotItem, [
            #     (PlotDataItem, tracedata, [
            #       ('selector1key', 'selector1value'),
            #       ('selector2key', 'selector2value'), ... more data selection criteria
            #     ]), ... more traces inside this plot
            #   ), ... more plot panels
            # ]
            self.plots = []
            for pl in plotlayout:
                # Start with default plot, update with provided plot layout configuration
                p = self._get_plot_defaults()
                p.update(pl)
                plot = self.addPlot(bottom="Time (s)", left=p["ylabel"], enableMenu=False)
                plot.setAxisItems({"bottom": pg.DateAxisItem()})
                plot.showGrid(x=True, y=True)
                plot.addLegend(horSpacing=25)
                if xlink and len(self.plots) > 0:
                    plot.setXLink(self.plots[0][0])
                # Add the traces to this plot panel
                traces = []
                for t in p["traces"]:
                    traceaxis = np.full(npoints, np.nan)
                    tracedata = np.full(npoints, np.nan)
                    # Remove selector from dictionary, pass key/value pairs as params to plot init method
                    selector = t.pop("selector")
                    trace = plot.plot(**t)
                    traces.append((trace, traceaxis, tracedata, selector))
                # Add interactive mouse crosshair
                vline = pg.InfiniteLine(angle=90, movable=False)
                hline = pg.InfiniteLine(angle=0, movable=False, label="{value:g}", labelOpts={"position": 1.0, "anchors": [(1.0, 0.0), (1.0, 1.0)]})
                #hline = pg.InfiniteLine(angle=0, movable=False, label="{value:g}", labelOpts={"position": 0.0, "anchors": [(0.0, 0.0), (0.0, 1.0)]})
                vline.hide()
                hline.hide()
                plot.addItem(vline, ignoreBounds=True)
                plot.addItem(hline, ignoreBounds=True)
                # Add plot to the plot/traces data structure
                self.plots.append((plot, traces, (hline, vline)))
                self.nextRow()
            # Connect to mouse move event to update crosshair position
            if crosshair == True and len(self.plots) > 0:
                self.plots[0][0].scene().sigMouseMoved.connect(self._mouse_moved)
                # Capture mouse leave events to hide crosshairs
                self.installEventFilter(self)

            if not self.data_queue is None:
                self.timer = QtCore.QTimer()
                self.timer.timeout.connect(self._update_plot)
                self.timer.start(15)


        def _mouse_moved(self, pos):
            for plot, traces, (hline, vline) in self.plots:
                legend = plot.addLegend()
                if plot.sceneBoundingRect().contains(pos):
                    point = plot.getViewBox().mapSceneToView(pos)
                    vline.show()
                    hline.show()
                    vline.setPos(point.x())
                    hline.setPos(point.y())
                    for trace, traceaxis, tracedata, _ in traces:
                        try:
                            y = tracedata[np.nanargmin(np.abs(traceaxis - point.x()))]
                            legend.getLabel(trace).setText(f"{trace.name()} = {y:g}")
                        except:
                            pass
                else:
                    vline.hide()
                    hline.hide()
                    for trace, _, _, _ in traces:
                        legend.getLabel(trace).setText(f"{trace.name()}")
                legend.updateSize()


        def eventFilter(self, obj, event):
            if obj == self:
                if event.type() == QtCore.QEvent.Type.Leave:
                    # Hide crosshair when mouse leaves the plot window
                    for plot, traces, (hline, vline) in self.plots:
                        legend = plot.addLegend()
                        vline.hide()
                        hline.hide()
                        for trace, _, _, _ in traces:
                            legend.getLabel(trace).setText(f"{trace.name()}")
                        legend.updateSize()
            return False


        def _get_plot_defaults(self):
            return {
                "ylabel": "Value (a.u.)",
                "traces": [{
                    "name": "Trace 1",
                    "pen": (255, 255, 255),
                    "selector": [
                        ["type", ".*"],
                    ]
                }]
            }


        def _update_plot(self):
            while not self.data_queue.empty():
                try:
                    data = self.data_queue.get(timeout=0.05)
                except queue.Empty:
                    return
                for d in data:
                    # Look through plot trace selector criteria looking for matching data
                    for p in self.plots:
                        for trace, traceaxis, tracedata, selectors in p[1]:
                            # Start by assuming data matches the selection criteria
                            match = True
                            try:
                                # Loop through each selection criteria
                                for s_k, s_v in selectors:
                                    # Try looking for this key in this data entry
                                    v = d[s_k]
                                    # This key exists, check its value
                                    if s_v == v or ((type(s_v) == type(v) == str) and re.fullmatch(s_v, v)) or s_v is NotImplemented:
                                        # Value matches this trace so far, keep checking any remaining selection criteria
                                        continue
                                    else:
                                        # Value doesn't match this trace, can stop checking now
                                        match = False
                                        break
                            except (IndexError, KeyError, ValueError) as ex:
                                # An exception means we couldn't match this data entry to this trace
                                print(ex)
                                match = False
                            if match and "value" in d:
                                # This data entry matched the selection criteria for this trace, update it
                                tracedata[0:-1] = tracedata[1:]
                                tracedata[-1] = d["value"]
                                # Get timestamp from data, or create one if it doesn't exist
                                traceaxis[0:-1] = traceaxis[1:]
                                try:
                                    traceaxis[-1] = d["timestamp"].timestamp()
                                except:
                                    traceaxis[-1] = datetime.now().timestamp()
                                trace.setData(x=traceaxis, y=tracedata)





"""
The datalogd package contains the main :class:`DataLogDaemon`,
plus the plugin base classes :class:`DataSource`, :class:`DataFilter`, and
:class:`DataSink`, which must be extended to provide useful functionality.

The included data source/filter/sink plugins are contained separately in the
:mod:`~datalogd.plugins` package.
"""

import os
import sys
import logging
import asyncio
import json
import argparse
import configparser

import pluginlib
import pydot
import appdirs

class DataLogDaemon():
    """
    The main datalogd class.

    The :class:`~datalogd.DataLogDaemon` reads configuration file(s), interprets
    the connection graph DOT specification, and initialises data
    source/filter/sink plugins and connections. The :mod:`asyncio` event loop
    must be started separately. For an example of this, see the
    :meth:`~datalogd.main` method, which is the typical way the daemon is
    started.

    :param configfile: Path to configuration file to load.
    :param plugindirs: Directory, or list of directories from which to load
        additional plugins.
    :param graph_dot: Connection graph specified in the DOT graph description
        language.
    """

    def __init__(self, configfile=None, plugindirs=[], graph_dot=None):
        self.log = logging.getLogger("DataLogDaemon")
        appname = "datalogd"

        # Set up default configuration
        config = configparser.ConfigParser()
        config.read_dict({
            f"{appname}": {
                "connection_graph": "digraph {\n  # default/empty graph, specify one in config file or command line parameter\n}",
                "plugin_paths": "[]",
            }
        })
        # Attempt to update configuration from files

        conf_files = [
            os.path.join(appdirs.site_config_dir(appname, False), f"{appname}.conf"),
            os.path.join(appdirs.user_config_dir(appname, False), f"{appname}.conf"),
        ]
        if configfile: conf_files.append(configfile)
        read_files = config.read(conf_files)
        if read_files:
            self.log.info(f"Loaded config from: {', '.join(read_files)}")
        else:
            self.log.log(logging.INFO if graph_dot else logging.WARNING,
                "No configuration file(s) were loaded. Looked for:\n  {}".format('\n  '.join(conf_files)))

        # Create list of plugin paths
        plugin_paths = listify(parse_dot_json(config.get(appname, "plugin_paths")))
        if plugindirs: plugin_paths.extend(plugindirs)

        # Show lists of available plugins
        try:
            loader = pluginlib.PluginLoader(group="datalogd", library="datalogd", paths=plugin_paths)
        except Exception as ex:
            raise RuntimeError(f"Error loading plugins: {ex.__name__}: {ex}")

        self.log.info(f"Detected source plugins: {', '.join(list(loader.plugins['DataSource']))}")
        self.log.info(f"Detected filter plugins: {', '.join(list(loader.plugins['DataFilter']))}")
        self.log.info(f"Detected sink plugins: {', '.join(list(loader.plugins['DataSink']))}")

        # Attempt to load connection graph=
        dot = graph_dot if graph_dot else config.get(appname, "connection_graph").strip()
        try:
            (graph,) = pydot.graph_from_dot_data(dot)
        except Exception as ex:
            raise RuntimeError(f"Unable to interpret DOT connection graph:\n{dot}")

        # Initialise specified nodes
        self.nodes = {}
        for n in graph.get_nodes():
            # Work around some issue with newline being interpreted as a node called "\n"
            if n.get_name() == "\"\\n\"":
                continue
            try:
                nodeclass = n.get_attributes()["class"]
            except KeyError:
                raise RuntimeError(f"Error reading DOT connection graph. Node {n.get_name()} does not have a \"class\" attribute.")

            # Gather contructor args and kw args from atribute string representations
            attributes = n.get_attributes()
            attributes.pop("class")
            args = []
            kwargs = {}
            for k, v in attributes.items():
                if v is None:
                    args.append(parse_dot_json(k))
                else:
                    kwargs[k] = parse_dot_json(v)
            self.log.info("Initialising node {}:{}({}{}{})".format(
                n.get_name(),
                nodeclass,
                ", ".join(args),
                ", " if args else "",
                ", ".join([f"{k}={v}" for k, v in kwargs.items()])))

            # Find the plugin from class name and initialise it
            try:
                nc = loader.get_plugin("Data" + nodeclass.rpartition("Data")[2], nodeclass)
                if nc is None:
                    raise RuntimeError(f"Error reading DOT connection graph. Can't find plugin for class {n.get_name()}:{nodeclass}.")
                self.nodes[n.get_name()] = nc(*args, **kwargs)
            except Exception as ex:
                raise RuntimeError(f"Unable to initialse node {n.get_name()}:{nodeclass} using args={args}, kwargs={kwargs}\n{ex}")

        # Ensure at least one node was initialised
        if not self.nodes:
            raise RuntimeError(f"No plugin nodes were initialised. Connection graph was:\n{dot}")

        # Make data connections from graph edges
        for e in graph.get_edges():
            try:
                self.log.info(f"Connecting {e.get_source()}:{type(self.nodes[e.get_source()]).__name__} -> {e.get_destination()}:{type(self.nodes[e.get_destination()]).__name__}")
                self.nodes[e.get_source()].connect_sinks(self.nodes[e.get_destination()])
            except KeyError:
                self.log.warning(f"Can't connect {e.get_source()} -> {e.get_destination()}, check node names are correct.")

    def close(self):
        """
        Notify nodes that the application is closing so they may shutdown gracefully.
        """
        for n in self.nodes.values():
            n.close()


def main():
    """
    Read command line parameters, instantiate a new :class:`DataLogDaemon` and
    begin execution of the event loop.
    """
    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger("main")
    # Read command line parameters
    argparser = argparse.ArgumentParser(description="Run the data logging daemon service.")
    argparser.add_argument("-c", "--configfile", help="Path to configuration file.", metavar="FILE")
    argparser.add_argument("-p", "--plugindirs", help="Directories containing additional plugins.", metavar="DIR", nargs="+")
    argparser.add_argument("-g", "--graph-dot", help="Connection graph specified in DOT format.")
    argparser.add_argument("--show-config-dirs", help="Display the default locations of configuration files, then exit.", action="store_true")
    args = argparser.parse_args()

    if args.show_config_dirs:
        log.info("Default configuration file locations are:\n  {}\n  {}".format(
            os.path.join(appdirs.site_config_dir("datalogd", False), "datalogd.conf"),
            os.path.join(appdirs.user_config_dir("datalogd", False), "datalogd.conf")))
        sys.exit(0)
    del args.show_config_dirs

    log.info("Initialising DataLogDaemon.")
    try:
        dld = DataLogDaemon(**vars(args))
    except Exception as ex:
        log.error(ex)
        sys.exit(1)
    # Start up the event loop to begin handling data flows
    log.info("Starting event loop.")
    loop = asyncio.get_event_loop()
    try:
        loop.run_forever()
    except KeyboardInterrupt as ex:
        pass
    log.info("Stopping event loop.")
    loop.close()
    dld.close()


if __name__ == "__main__":
    main()


def listify(value):
    """
    Convert ``value`` into a list.

    Modifies the behaviour of the python builtin :meth:`list` by accepting all
    types as ``value``, not just iterables. Additionally, the behaviour of
    iterables is changed:

    * ``list('str') == ['s', 't', 'r']``, while ``listify('str') == ['str']``
    * ``list({'key': 'value'}) == ['key']``, while ``listify({'key': 'value'}) == [{'key': 'value'}]``

    :param value: Input value to convert to a list.
    :returns: ``value`` as a list.
    """
    if type(value) == str: return [value]
    if type(value) == dict: return [value]
    try:
        value = list(value)
    except TypeError:
        value = [value]
    return value


def parse_dot_json(value):
    """
    Interpret the value of a DOT attribute as JSON data.

    DOT syntax requires double quotes around values which contain DOT
    punctuation (space, comma, {}, [] etc), and, if used, these quotes will also
    be present in the obtained value string. Unfortunately, JSON also uses
    double quotes for string values, which are then in conflict. This method
    will strip any double quotes from the passed ``value``, then will attempt to
    interpret as JSON after replacing single quotes with double quotes.

    Note that the use of this workaround means that single quotes must be used
    in any JSON data contained in the DOT attribute values.

    Although not strictly correct JSON, some special values will be interpreted as their python
    equivalents. These are:

      - None or null (with any capitalisation) will be read as a python ``None``.
      - True (with any capitalisation) will be read as a python ``True``.
      - False (with any capitalisation) will be read as a python ``False``.
      - NotImplemented (with any capitalisation) will be read as a python ``NotImplemented``.
      - NaN (with any capitalisation) will be read as the python float ``nan``.
      - Inf or Infinity (with any capitalisation) will be read as the python float ``inf``.
      - -Inf or -Infinity (with any capitalisation) will be read as the python float ``-inf``.

    :param value: string to interpret.
    :returns: ``value``, possibly as a new type.
    """
    # None, or unquoted versions of "None", "null" is None
    if value is None or value.strip().lower() in ("none", "null"): return None
    # Handle "True" and "False", not just "true" and "false"
    if value is True or value.lower() == "true": return True
    if value is False or value.lower() == "false": return False
    # Handle the python special value NotImplemented
    if value is NotImplemented or value.lower() == "NotImplemented": return NotImplemented
    # Handle all capitalisations of "NaN", and "+nan", "-NaN", " NAN"
    if value is float("nan") or value.strip("+- ").lower() == "nan": return float("nan")
    # Handle infinity and -infinity
    if value is float("inf") or value.strip("+ ").lower() in ("inf", "infinity"): return float("inf")
    if value is float("-inf") or value.strip().lower() in ("-inf", "-infinity"): return float("-inf")
    try:
        # Attempt to interpret as JSON
        # First strip any double quotes used to quote in DOT file
        if len(value) > 1 and value[0] == value[-1] == '"':
            value = value[1:-1]
        # Now replace single quotes with doubles to make correct JSON
        jvalue = value.replace("'", '"')
        value = json.loads(jvalue)
    except json.JSONDecodeError as ex:
        logging.debug(f"Unable to parse DOT value as JSON, falling back to string: {ex.msg}")
        # Fall back to string, strip any quotes used to force string mode
        if len(value) > 1 and value[0] == value[-1] and value[0] in "'\"":
            value = value[1:-1]
    return value


@pluginlib.Parent(group="datalogd")
class DataSource():
    """
    The base class for all data sink plugins.

    :class:`~datalogd.DataSource` implements methods for connecting or
    disconnecting sinks, and for sending data to connected sinks. It has no
    intrinsic functionality (it does not actually produce any data) and is not
    itself considered a plugin, so can't be instantiated using the connection
    graph.

    :param sinks: :class:`~datalogd.DataSink` or list of
        :class:`~datalogd.DataSink`\ s to receive data produced by this
        :class:`~datalogd.DataSource`.
    """
    def __init__(self, sinks=[]):
        self.sinks = []
        if sinks: self.connect_sinks(sinks)
        self.log = logging.getLogger("DataSource")

    def send(self, data):
        """
        Send the provided ``data`` to all connected :class:`DataSink`\ s.

        :param data: Data to send to :class:`DataSink`\ s.
        """
        for s in self.sinks:
            s.receive(data)

    def connect_sinks(self, sinks):
        """
        Register the provided :class:`DataSink` as a receiver of data produced
        by this :class:`~datalogd.DataSource`. A list of sinks may also be
        provided.

        :param sinks: :class:`~datalogd.DataSink` or list of
            :class:`~datalogd.DataSink`\ s.
        """
        sinks = listify(sinks)
        for s in sinks:
            if s in self.sinks:
                self.log.warning(f"Not adding duplicated sink: {s}")
            if s is self:
                self.log.warning(f"Not adding self as own sink: {s}")
            else:
                try:
                    if callable(s.receive):
                        self.sinks.append(s)
                    else:
                        raise AttributeError
                except AttributeError:
                    self.log.warning(f"Skipping invalid sink: The {type(s).__name__} \"{s}\" does not have a receive() method.")

    def disconnect_sinks(self, sinks):
        """
        Unregister the provided :class:`~datalogd.DataSink` so that it no longer
        receives data produced by this :class:`~datalogd.DataSource`. A list of
        sinks may also be provided. It is not an error to provide a sink that is
        not currently connected.

        :param sinks: :class:`~datalogd.DataSink` or list of
            :class:`~datalogd.DataSink`\ s.
        """
        sinks = listify(sinks)
        for s in sinks:
            for ss in self.sinks:
                if ss is s:
                    self.sinks.remove(s)

    def close(self):
        """
        Perform any cleanup required during application shutdown.
        """
        pass


class NullDataSource(DataSource):
    """
    A :class:`~datalogd.DataSource` which produces no data.

    Unlike the base :class:`~datalogd.DataSource`, this can be instantiated
    using the connection graph, although it provides no additional
    functionality.
    """


@pluginlib.Parent(group="datalogd")
class DataSink():
    """
    The base class for all data sink plugins.

    :class:`~datalogd.DataSink`\ s have a :meth:`receive` method which accepts
    data from connected :class:`~datalogd.DataSource`\ s.
    """
    def receive(self, data):
        """
        Accept the provided ``data``.

        :param data: Data received by this sink.
        """

    def close(self):
        """
        Perform any cleanup required during application shutdown.
        """
        pass


class NullDataSink(DataSink):
    """
    A :class:`~datalogd.DataSink` which accepts data and does nothing with it.

    Unlike the base :class:`~datalogd.DataSink`, this can be instantiated using
    the connection graph, although it provides no additional functionality.
    """


@pluginlib.Parent(group="datalogd")
class DataFilter(DataSource, DataSink):
    """
    The base class for all data filter plugins.

    :class:`~datalogd.DataFilter`\ s are subclasses of both
    :class:`~datalogd.DataSource`\ s and :class:`~datalogd.DataSink`\ s, thus
    are capable of both sending and receiving data. Typically, they are used to
    sit between a :class:`~datalogd.DataSource` and a
    :class:`~datalogd.DataSink` (or other :class:`~datalogd.DataFilter`\ s) in
    order to modify the data flowing between them in some way.
    """


class NullDataFilter(DataFilter):
    """
    A :class:`~datalogd.DataFilter` which accepts data and passes it unchanged
    to any connected :class:`~datalogd.DataSink`\ s.
    """
    def receive(self, data):
        """
        Pass ``data`` unchanged to all connected :class:`~datalogd.DataSink`\ s.

        :param data: Data received by this filter.
        """
        self.send(data)



from datalogd import DataFilter, listify

class JoinDataFilter(DataFilter):
    """
    Join two or more consecutive receipts of data together into a list.

    If the data are already lists, the two lists will be merged.

    :param count: Number of data receipts to join.
    """
    def __init__(self, sinks=[], count=2):
        super().__init__(sinks=sinks)
        self.count = count
        self._buffer = []

    def receive(self, data):
        """
        Accept ``data`` and join consecutive receipts of data together into a list.

        :param data: data to join.
        """
        data = listify(data)
        self._buffer.append(data)
        if len(self._buffer) >= self.count:
            joined_data = []
            for b in self._buffer:
                joined_data.extend(b)
            self.send(joined_data)
            self._buffer = []



import asyncio
import logging
import random

from datalogd import DataSource

class RandomWalkDataSource(DataSource):
    """
    Generate test or demonstration data using a random walk algorithm.

    For each iteration of the algorithm, the output value will either be
    unchanged, increase, or decrease by a fixed increment. The options are
    chosen randomly with equal probability.

    Multiple walkers can be initialised to produce several sources of random
    data. The ``walkers`` parameter is a list, the length of which determines
    the number of walkers to use. Each item in the list must be a list/tuple of
    two items: the walker's initial value and increment.

    :param seed: Seed used to initialise the random number generator.
    :param interval: How often to run an iteration of the algorithm, in seconds.
    :param walkers: List defining number of walkers and their parameters in the
        form ``[[init, increment], ...]``.
    """
    def __init__(self, sinks=[], seed=None, interval=1.0, walkers=[[0.0, 1.0], [0.0, 2.0]]):
        super().__init__(sinks=sinks)
        self.log = logging.getLogger("RandomWalkDataSource")
        random.seed(seed)
        self.interval = interval
        self.walkers = []
        for w in walkers:
            try:
                self.walkers.append([float(w[0]), float(w[1])])
            except Exception as ex:
                self.log.warning(f"Invalid parameters for walker: {w} {ex}")
        # Queue first call of update routine
        asyncio.get_event_loop().call_soon(self.generate_data)

    def generate_data(self):
        """
        Run one iteration of the random walk algorithm and send the value to any
        connected sinks.
        """
        loop = asyncio.get_event_loop()
        # Update walker values and notify sinks
        data = []
        for w_i, w in enumerate(self.walkers):
            w[0] += random.choice([-w[1], 0.0, w[1]])
            data.append({"type": "randomwalk", "id": f"{w_i}", "value": w[0]})
        self.send(data)
        # Reschedule next update
        loop.call_later(self.interval, self.generate_data)



"""
The plugins package contains the included :class:`DataSource <datalogd.DataSource>`,
:class:`DataFilter <datalogd.DataFilter>`, and :class:`DataSink <datalogd.DataSink>`
subclasses.
Some plugins will require additional python modules to be installed, or may have
other specific requirements (for example, the :class:`LibSensorsDataSource
<datalogd.plugins.libsensors_datasource.LibSensorsDataSource>` probably won't work on Windows
operating systems).
"""



import logging
from datetime import datetime, timezone

from datalogd import DataSink, listify

try:
    from influxdb import InfluxDBClient
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("influxdb module not found. To use the InfluxDBDataSink, install the module with \"pip install influxdb\" or similar.")
else:
    # Required modules present, continue loading rest of this module

    class InfluxDBDataSink(DataSink):
        """
        Connection to a InfluxDB database for storing time-series data.

        .. Note::

            This plugin is outdated. For new InfluxDB 2.x installs, use the
            :class:`~datalogd.plugins.influxdb2_datasink.InfluxDB2DataSink` plugin instead.

        Note that this doesn't actually run the InfluxDB database service, but simply connects to an
        existing InfluxDB database via a network (or localhost) connection. See the `getting started
        <https://v2.docs.influxdata.com/v2.0/get-started/>`_ documentation for details on
        configuring a new database server.

        :param host: Host name or IP address of InfluxDB server.
        :param port: Port used by InfluxDB server.
        :param user: Name of database user.
        :param password: Password for database user.
        :param dbname: Name of database in which to store data.
        :param session: A name for the measurement session.
        :param run: A tag to identify commits from this run. Default of ``None``
            will use a date/time stamp.
        """
        def __init__(self, host="localhost", port=8086, user="root", password="root", dbname="datalogd", session="default", run=None):
            self.log = logging.getLogger("InfluxDBDataSink")
            try:
                self.client = InfluxDBClient(host, port, user, password, dbname)
            except Exception as Ex:
                self.log.warning("Unable to make connection to InfluxDB database.")
            self.session = session
            if run is None:
                self.run = datetime.now(timezone.utc).astimezone().strftime("%Y%m%d-%H%M%S")
            else:
                self.run = run

        def receive(self, data):
            """
            Commit data to the InfluxDB database.

            Multiple items of data can be submitted at once if ``data`` is a list. A typical format
            of ``data`` would be::

                [
                    {'type': 'temperature', 'id': '0', 'value': 22.35},
                    {'type': 'humidity', 'id': '0', 'value': 55.0},
                    {'type': 'temperature', 'id': '1', 'value': 25.80},
                ]

            The data point will have its data field generated using the form ``<type>_<id> =
            <value>``.

            A timestamp for the commit will be generated using the current system clock if a
            "timestamp" field does not already exist.

            :param data: Data to commit to the database.
            """
            if data is None or data == []: return
            data = listify(data)

            # Start building the structure to enter into database
            datapoints = {
                "measurement": self.session,
                "tags": {"run": self.run},
                "fields": {},
            }
            for d in data:
                if "type" in d.keys():
                    k = str(d["type"]) + ("_" + str(d["id"])) if "id" in d.keys() else ""
                    v = d["value"] if "value" in d.keys() else None
                else:
                    k = str(d)
                    v = None
                datapoints["fields"][k] = v
                if "timestamp" in d.keys():
                    if type(d["timestamp"]) == datetime:
                        datapoints["time"] = d["timestamp"].isoformat()
                    else:
                        datapoints["time"] = d["timestamp"]
                else:
                    datapoints["time"] =  datetime.now(timezone.utc).astimezone().isoformat()
            # Send structure out to database
            try:
                self.log.debug(f"Comitting: {datapoints}")
                self.client.write_points([datapoints])
            except Exception as ex:
                self.log.warning("Unable to commit data to InfluxDB database.", exc_info=True)

        def close(self):
            """
            Close the connection to the database.
            """
            self.client.close()


import asyncio
import logging

import json

from datalogd import parse_dot_json
from datalogd import DataSource

try:
    import serial, serial.tools.list_ports, serial.threaded, serial_asyncio
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("Serial modules not found. Install with \"pip install pyserial pyserial-asyncio\" or similar.")
else:
    # Required modules present, continue loading rest of this module

    class SerialDataSource(DataSource):
        """
        Receive data from an Arduino connected via a serial port device.

        .. container:: toggle

            .. container:: header

                See the ``datalog_arduino.ino`` sketch for matching code to run
                on a USB-connected Arduino.

            .. literalinclude:: ../../../arduino/datalog/datalog.ino
                :language: c++
                :caption: ``datalog.ino``

        Other serial-connected devices should work with this class if they conform to the expected
        communications protocol. Message data should be encoded in a JSON format. For example

        .. code-block::

            {"board":"A","timestamp":"1601251","message":"measurement","data":[{"type":"temperature","source":"A","id":"A_TC0","value":"20.25","units":"C"}]}

        which describes a single temperature measurement data point, encapsulated by a message
        header. Note that the values encoded in the ``"value"`` field will be attempted to be
        decoded using the same logic as :data:`~datalogd.parse_dot_json`, so that ``"20.25"`` will
        be interpreted as the equivalent python float, and special values such as ``None`` and
        ``inf`` are supported.

        If the connection to the serial device cannot be established or is interrupted, regular
        reattempts will be performed. Note that this means an exception will not be raised if the
        serial device cannot be found.

        :param port:     Path of serial device to use. A partial name to match
            can also be provided, such as "usb".
        :param board_id: ID label provided by the Arduino data logging board, to
            select a particular device in case multiple boards are connected.
        """

        class SerialHandler(serial.threaded.LineReader):
            """
            A class used as a :mod:`asyncio` :class:`~asyncio.Protocol` to handle
            lines of text received from the serial device.

            :param parent: The parent :class:`~datalogd.plugins.serial_datasource.SerialDataSource` class.
            """
            def __init__(self, parent):
                super().__init__()
                self.parent = parent

            def handle_line(self, line):
                """
                Accept one line of text, parse it to extract data, and pass the
                data on to any connected sinks.

                :param line: Line of text to process.
                """
                try:
                    j = json.loads(line)
                    if j["message"] == "measurement":
                        self.parent.log.debug(f"Received: {j['data']}")
                        # All data is in string form, attempt to convert values to something more appropriate
                        try:
                            for d in j["data"]:
                                if "value" in d.keys():
                                    d["value"] = parse_dot_json(d["value"])
                        except Exception as ex:
                            self.parent.log.warning("Unable to parse serial data.", exc_info=True)
                        self.parent.send(j["data"])
                except Exception as ex:
                    raise RuntimeError(f"Error interpreting serial data: {ex}")

            def connection_lost(self, exc):
                self.parent.log.warning("Serial connection lost, will attempt to reconnect.")
                self.parent._connection_task = self.parent.loop.create_task(self.parent._connect_serial_port())


        def __init__(self, sinks=[], port="", board_id=None):
            super().__init__(sinks=sinks)
            self.log = logging.getLogger("SerialDataSource")
            self.port = port
            self.sp = None
            self.board_id = board_id
            # Get reference to event loop and schedule task (but loop probably isn't started yet)
            self.loop = asyncio.get_event_loop()
            self._connection_task = self.loop.create_task(self._connect_serial_port(), name="SerialDataSource Connector")


        async def _connect_serial_port(self):
            """
            Coroutine to attempt to find and connect to device over serial port.
            """
            while True:
                # Loop continuously attempting to connect to correct serial device
                self.sp = None
                # Get list of available serial ports (matching given port name)
                portlist = list(serial.tools.list_ports.grep(self.port))
                if len(portlist) == 0:
                    if self.port == "":
                        self.log.warning("No serial ports found.")
                    else:
                        self.log.warning(f"No serial ports found matching \"{self.port}\"")
                
                # Iterate though serial ports looking for requested logging board
                for p in portlist:
                    try:
                        self.sp = serial.Serial(p.device, 115200, timeout=2)
                        # Read and discard potentially partial line
                        self.sp.readline()
                        # Read and attempt json decode of next (complete) line
                        j = json.loads(self.sp.readline().decode("ascii").strip())
                        if j["board"] and j["timestamp"] and j["message"]:
                            # Looks like one of our logging boards
                            if self.board_id is None or j["board"] == str(self.board_id):
                                self.log.info(f"Found board \"{j['board']}\" at {p.device}")
                                break
                            else:
                                self.log.info(f"Found board \"{j['board']}\" at {p.device} (but is not requested board \"{self.board_id}\")")
                                self.sp.close()
                                self.sp = None
                                continue
                    except Exception as ex:
                        # Error opening serial device, or not a logging board
                        self.log.info(f"Error querying board at {p.device} (port error or received invalid data)")
                        try:
                            self.sp.close()
                        except:
                            pass
                        self.sp = None

                # We should now have opened the serial port to requested board...
                if self.sp is None:
                    # ...but failed, so try again in a little while
                    try:
                        await asyncio.sleep(5.0)
                    except asyncio.CancelledError:
                        self.log.info("Serial connection attempt cancelled.")
                        break
                else:
                    # All seems OK, start the serial reader coroutine and stop connection attempts
                    asyncio.create_task(self._create_reader_coroutine(), name="SerialDataSource Reader")
                    break


        async def _create_reader_coroutine(self):
            """
            Coroutine to create the serial port transport and protocol class instances.
            """
            protocol = self.SerialHandler(parent=self)
            transport = serial_asyncio.SerialTransport(self.loop, protocol, self.sp)
            return (transport, protocol)


        def close(self):
            """
            Close the serial port connection.
            """
            if not self.sp is None:
                try:
                    self.sp.close()
                finally:
                    self.sp = None



from datetime import datetime, timezone

from datalogd import DataFilter, listify

class TimeStampDataFilter(DataFilter):
    """
    Add or update a timestamp field to data using the current system clock.
    """
    def receive(self, data):
        """
        Accept the provided ``data`` and add a timestamp field.

        If ``data``, or elements in the ``data`` list are ``dict``\ s, then a
        "timestamp" field will be added. Otherwise, the data entries will be
        converted to a ``dict`` with the old entry stored under a "value" field.

        :param data: Data to add a timestamp to.
        """
        data = listify(data)
        ts_data = []
        # Use local time, but ensure timezone information is present
        timestamp = datetime.now(timezone.utc).astimezone()
        for d in data:
            if isinstance(d, dict):
                ts_data.append(d)
            else:
                ts_data.append({"value": d})
            ts_data[-1]["timestamp"] = timestamp
        self.send(ts_data)



import re

import numpy as np

from datalogd import DataFilter, listify

class PolynomialFunctionDataFilter(DataFilter):
    """
    Select data based on key--value pairs, then apply a polynomial function to a value.

    Any data which matches all of the ``match_keyvals`` key--value pairs will be processed. The
    format of the ``match_keyvals`` parameter is a list in the form ``[[key, value], [key2,
    value2]...]``. For example, ``match_keyvals=[["type", "temperature"], ["id", "123"]]`` will
    process any data which has a ``"type"`` field of ``"temperature"`` and an ``"id"`` field of
    ``"123"``. A ``value`` of the python special ``NotImplemented`` will match any value for the
    given key. In the case that values are strings, they will be matched as regular expressions, for
    example ``".*"`` will match any string.

    Once a data item is matched, a value will be selected to apply the polynomial function to,
    selected by the ``value`` parameter. By default this is the value stored under the ``"value"`` key.

    The polynomial function is defined by a set of coefficients, given by the ``coeffs`` parameter.
    This is an array of :math:`n` coefficients, :math:`c_n`, which forms the function
    :math:`x^\prime = \sum_n c_n x^{(n-1)} \equiv c_0 + c_1x + c_2x^2 \ldots c_nx^n`. For example,
    ``coeffs=[1.23, 1.0]`` would add 1.23 to a value, while ``coeffs=[0, 10]`` would multiply a
    value by 10. Specifying additional coefficients include quadratic, cubic terms etc.

    Rounding may be applied to the result by supplying the number of decimal places in the
    ``rounding`` parameter. Rounding behaviour is determined by the numpy ``around()`` function.
    Negative numbers specify positions to the left of the decimal point.

    The value of the data entry's ``"units"`` field can be modified or created using the ``units``
    parameter. For example, ``units="V"`` might be used to indicate that an analogue measurement in
    arbitrary units now equates to voltage, determined by the polynomial function calibration curve.

    :param match_keyvals: Key--value pairs to match to data items.
    :param value: Key from data item containing the value to modify.
    :param coeffs: Coefficients of the polynomial function to apply.
    :param rounding: Number of decimal places to round the result.
    :param units: New value of units field for the modified data item.
    """
    def __init__(self, sinks=[], match_keyvals=[["type", ".*"], ["id", ".*"]], value="value", coeffs=[0.0, 1.0], rounding=None, units=None):
        super().__init__(sinks=sinks)
        self._keyvals = match_keyvals
        self._value = value
        self._coeffs = coeffs
        self._rounding = rounding
        self._units = units


    def receive(self, data):
        """
        Accept the provided ``data``, select based on key/value pairs, apply function, and pass
        onto connected sinks.

        The selection is based upon the parameters provided to the constructor of this
        :class:`~datalogd.plugins.polynomialfunction_datafilter.PolynomialFunctionDataFilter`.

        :param data: Data to correct.
        """
        data = listify(data)
        for d in data:
            try:
                # Look for matches to all key/value pairs
                match = True
                for kv_k, kv_v in self._keyvals:
                    # Try looking for this key in this data entry
                    v = d[kv_k]
                    # This key exists, check its value
                    if kv_v == v or ((type(kv_v) == type(v) == str) and re.fullmatch(kv_v, v)) or kv_v is NotImplemented:
                        # Value matches, keep checking any remaining keyval pairs
                        continue
                    else:
                        # Value doesn't match, can stop checking now
                        match = False
                        break
                if match:
                    # Retrieve specified value to correct
                    x = d[self._value]
                    # Allow correcting numpy arrays
                    if isinstance(x, np.ndarray):
                        y = np.zeros_like(x)
                    else:
                        y = 0
                    # Apply polynomial function
                    for n, c in enumerate(self._coeffs):
                        y += c*x**n
                    # Apply rounding if requested
                    if self._rounding is not None:
                        y = np.round(y, int(self._rounding))
                    d[self._value] = y
                    # Change or add new units to the data if given
                    if self._units is not None:
                        d["units"] = self._units
            except (IndexError, KeyError, ValueError):
                # An exception means we couldn't match this data entry, or find specified value
                pass
        self.send(data)





from os import path
from datetime import datetime, timezone
import asyncio

from datalogd import DataSink

class FileDataSink(DataSink):
    """
    Write data to a file.

    By default, any existing contents of ``filename`` will be overwritten without
    prompting. To instead raise an error if the file exists, set the ``mode``
    parameter to 'x'. The contents of any existing file will be appended to by
    setting ``mode='a'``.

    To automatically prepend a date and time stamp to the given filename, set
    ``filename_timestamp=True``. In general, this should create a unique filename and prevent
    overwriting when ``filename`` already exists.

    The ``flush_interval`` parameter controls the behaviour of the file writes.
    It describes how often, in seconds, the operating system's buffers should be
    flushed to disk, updating the file contents:

        * ``flush_interval > 0`` causes the flush to occur at the given time
          interval, in seconds. More frequent flushes will keep the contents of
          the file updated, but put more strain on the machines I/O systems.
        * ``flush_interval = 0`` will flush immediately after each receipt of
          data.
        * ``flush_interval < 0`` will not automatically flush, leaving this to
          the operating system. The contents of the file may not update until
          the program closes.
        * ``flush_interval == None`` will perform a file open, write, and close
          operation on each receipt of data. This may be desired if the contents
          of the file should only contain the latest received data (and should
          be used in conjunction with the ``mode='w'`` parameter).


    :param filename: File name to write data to.
    :param filename_timestamp: Prepend a timestamp to the filename.
    :param mode: Mode in which to open the file. One of 'w' (write), 'a'
        (append), 'x' (exclusive creation).
    :param header: Header to write to file after plugin initialisation.
    :param terminator: Separator written to file after each receipt of data.
    :param flush_interval: Interval, in seconds, between flushes to disk.
    """
    def __init__(self, filename, filename_timestamp=False, mode="w", header="", terminator="\n", flush_interval=10):
        self.filename = filename
        self.filename_timestamp = bool(filename_timestamp)
        if self.filename_timestamp:
            (head, tail) = path.split(self.filename)
            tail = f"{datetime.now(timezone.utc).astimezone().strftime('%Y%m%d-%H%M%S')}-{tail}"
            self.filename = path.join(head, tail)
        self.mode = mode if mode in "wax" else "w"
        self.header = header
        self.terminator = terminator
        self.flush_interval = flush_interval
        if self.flush_interval is not None:
            # Keep file open, flush when data is ready
            try:
                self.fd = open(self.filename, self.mode)
                if self.header: self.fd.write(self.header)
            except Exception as ex:
                raise RunTimeError(f"Unable to write to file {filename}: {ex}")
        else:
            # Open and close file each time
            self.fd = None
        self._flush_handle = None

    def close(self):
        """
        Close the connection to the file.
        """
        if self._flush_handle: self._flush_handle.cancel()
        if self.fd: self.fd.close()

    def receive(self, data):
        """
        Accept ``data`` and write it out to the file.

        The ``terminator`` specified in the constructor will be appended to the file after each call
        of this method.

        :param data: Data to write to file.
        """
        if not self.fd:
            # Open and close file each time, (probably) wiping any old data
            try:
                fd = open(self.filename, self.mode)
                if self.header: fd.write(self.header)
                fd.write(data)
                fd.write(self.terminator)
            except Exception as ex:
                raise RunTimeError(f"Unable to write to file {filename}: {ex}")
        else:
            # Keeping file open, appending new data
            self.fd.write(data)
            self.fd.write(self.terminator)
            if self.flush_interval == 0:
                # Flush after every write
                self.fd.flush()
            elif self.flush_interval > 0:
                # Schedule a flush if one isn't already scheduled
                if not self._flush_handle:
                    self._flush_handle = asyncio.get_event_loop().call_later(self.flush_interval, self._flush)
            else:
                pass
                # Negative means don't flush (leave to operating system)

    def _flush(self):
        self.fd.flush()
        self._flush_handle = None



#!/usr/bin/env python3
import os
import sys
import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

systemd_unit = """[Unit]
Description=Data Logging Daemon

[Service]
Type=simple
ExecStart=/usr/bin/datalogd

[Install]
WantedBy=multi-user.target
"""


"""
Implements the distutils 'install' command to install service startup files.
"""
from setuptools.command.install import install
class CustomInstallCommand(install):

    def initialize_options(self):
        install.initialize_options(self)
        # Enable recording of installed files
        self.record = "installed_files.txt"
        self.file_list = []

    def run(self):
        if sys.platform.startswith("linux"):
            systemd_path = "/lib/systemd/system/"
            if os.access(systemd_path, os.W_OK):
                with open(os.path.join(systemd_path, "datalogd.service"), "w") as fd: fd.write(systemd_unit)
                self.file_list.append(os.path.join(systemd_path, "datalogd.service"))
                try:
                    os.system("systemctl daemon-reload")
                    os.system("systemctl enable datalogd")
                except Exception as ex:
                    print("Unable to communicate with systemctl.", ex)
                print("Installed systemd service file in system directory. Start it with \"systemctl start datalogd\"")
            else:
                try:
                    systemd_path = os.path.join(os.path.expanduser("~"), ".local/share/systemd/user")
                    os.makedirs(systemd_path, exist_ok=True)
                    with open(os.path.join(systemd_path, "datalogd.service"), "w") as fd:
                        fd.write(systemd_unit.replace("/usr/bin", os.path.join(os.path.expanduser("~"), ".local/bin")))
                        self.file_list.append(os.path.join(systemd_path, "datalogd.service"))
                    try:
                        os.system("systemctl --user daemon-reload")
                        os.system("systemctl --user enable datalogd")
                    except Exception as ex:
                        print("Unable to communicate with systemctl.", ex)
                    print("Installed systemd service file in user directory. Start it with \"systemctl --user start datalogd\"")
                except Exception as ex:
                    print("Unable to write to system-wide or user-wide systemd directories.", ex)

        elif sys.platform.startswith("win32"):
            print("Windows OS detected, requires manual startup configuration.")

        install.run(self)


    def get_outputs(self):
        """
        Append any custom install files to the file record list.
        """
        outputs = install.get_outputs(self)
        outputs.extend(self.file_list)
        return outputs


"""
Implements the distutils 'develop' command to install service startup files.
"""
from setuptools.command.develop import develop
class CustomDevelopCommand(develop):
    def run(self):
        if self.uninstall:
            print("TODO: Uninstall system service files.")
        else:
            print("TODO: Install system service files.")
        develop.run(self)


"""
Implements a custom distutils 'uninstall' command.
"""
from distutils.core import Command
class CustomUninstallCommand(Command):

    description = "Uninstall datalogd, including service startup files."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if sys.platform.startswith("linux"):
            # Stop any systemd services
            try:
                os.system("systemctl disable --now datalogd > /dev/null 2>&1")
            except: pass
            try:
                os.system("systemctl --user disable --now datalogd > /dev/null 2>&1")
            except: pass
        with open("installed_files.txt") as fd: filelist = fd.readlines()
        for f in filelist:
            try:
                os.remove(f.strip())
                print(f"Removed file: {f.strip()}")
            except Exception as ex:
                pass
            dirname = os.path.dirname(f.strip())
            if "datalogd" in dirname:
                try:
                    os.removedirs(dirname)
                    print(f"Removed directory: {dirname}")
                except Exception as ex:
                    pass
        if sys.platform.startswith("linux"):
            # Reload systemd units
            try:
                os.system("systemctl daemon-reload > /dev/null 2>&1")
            except: pass
            try:
                os.system("systemctl --user daemon-reload > /dev/null 2>&1")
            except: pass


setuptools.setup(
    name="datalogd",
    version="0.3.1",
    author="Patrick Tapping",
    author_email="mail@patricktapping.com",
    description="A data logging daemon, easily customisable using a flexible plugin system.",
    long_description=long_description,
    url="https://gitlab.com/ptapping/datalogd",
    project_urls={
        "Documentation": "https://datalogd.readthedocs.io/",
        "Source": "https://gitlab.com/ptapping/datalogd",
        "Tracker": "https://gitlab.com/ptapping/datalogd/-/issues",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pluginlib",
        "pydot",
        "appdirs",
    ],
    entry_points={
        "console_scripts": [
            "datalogd = datalogd:main",
        ],
    },
    cmdclass={
        'install': CustomInstallCommand,
        'develop': CustomDevelopCommand,
        'uninstall' : CustomUninstallCommand,
    },
)



import re

from datalogd import DataFilter, listify

class KeyValDataFilter(DataFilter):
    """
    Select or reject data based on key--value pairs.

    Received data items will be inspected to see whether they contain the given keys, and that their
    values match the given values. The key-value pairs are supplied as a list in the form ``[[key,
    value], [key2, value2]...]``. All key-value pairs must be matched. A value of the python
    special value of ``NotImplemented`` will match any value. If both ``value`` and ``data[key]`` are
    strings, matching will be performed using regular expressions (in which case ``".*"`` will match
    all strings). If the ``select`` flag is ``True``, only matching data will be passed on to the
    connected sinks, if it is ``False``, only non-matching data (or data that does not contain the
    given ``key``) will be passed on.

    If only a single key-value pair needs to be matched, they may alternatively be passed as the
    ``key`` and ``val`` parameters. This is mainly intended for backwards compatibility.

    :param select: Pass only matching data, or only non-matching data.
    :param keyvals: List of dictionary key-value pairs to match in incoming data.
    """
    def __init__(self, sinks=[], select=True, keyvals=None, key="type", val=None):
        super().__init__(sinks=sinks)
        self._select = bool(select)
        self._keyvals = keyvals
        if self._keyvals is None:
            self._keyvals = [[key, val]]


    def receive(self, data):
        """
        Accept the provided ``data``, and select or reject items before passing
        on to any connected sinks.

        The selection is based upon the parameters provided to the constructor
        of this :class:`~datalogd.plugins.keyval_datafilter.KeyValDataFilter`.

        :param data: Data to filter.
        """
        data = listify(data)
        data_accept = []
        for d in data:
            try:
                # Look for matches to all key/value pairs
                match = True
                for kv_k, kv_v in self._keyvals:
                    # Try looking for this key in this data entry
                    v = d[kv_k]
                    # This key exists, check its value
                    if kv_v == v or ((type(kv_v) == type(v) == str) and re.fullmatch(kv_v, v)) or kv_v is NotImplemented:
                        # Value matches, keep checking any remaining keyval pairs
                        continue
                    else:
                        # Value doesn't match, can stop checking now
                        match = False
                        break
            except (IndexError, KeyError, ValueError):
                # An exception means we couldn't match this data entry
                match = False
            if (match and self._select) or (not match and not self._select):
                data_accept.append(d)
        self.send(data_accept)



import asyncio
import logging
from enum import Enum

from datalogd import DataSource

try:
    import sensors
except ModuleNotFoundError:
    log = logging.getLogger(__name__.rpartition(".")[2])
    log.warning("sensors module not found. Install it with \"pip install PySensors\" or similar.")
else:
    # Required modules present, continue loading rest of this module

    class LibSensorsDataSource(DataSource):
        """
        Provide data about the running system's hardware obtained using the
        ``libsensors`` library.

        ``libsensors`` is present on most Linux systems, or can be installed
        from the distribution's repositories (``apt install libsensors5`` on
        Debian/Ubuntu, ``pacman -S lm_sensors`` on Arch etc.). The available
        sensors will depend on your hardware, Linux kernel, and version of
        ``libsensors``.

        Attempting to initialise this plugin on Windows operating systems will
        almost certianly fail.

        :param interval: How often to poll the sensors, in seconds.
        """
        def __init__(self, sinks=[], interval=1.0):
            super().__init__(sinks=sinks)
            self.interval = interval
            sensors.init()
            # Queue first call of update routine
            asyncio.get_event_loop().call_soon(self.read_sensors)

        def close(self):
            """
            Close the connection to the sensors.
            """
            sensors.cleanup()

        def read_sensors(self):
            """
            Read sensors and send data to any connected sinks.
            """
            loop = asyncio.get_event_loop()
            data = []
            for chip in sensors.iter_detected_chips():
                for feature in chip:
                    data.append({"type": f"{LibSensorsFeatureType(feature.type).type}", "source": "libsensors", "id": f"{chip}_{feature.label}", "value": feature.get_value()})
            self.send(data)
            # Reschedule next update
            loop.call_later(self.interval, self.read_sensors)


    class LibSensorsFeatureType(Enum):

        """ A utility :class:`~enum.Enum` used to interpret integers
        representing sensor feature types. """

        IN          = 0x00
        FAN         = 0x01
        TEMP        = 0x02
        POWER       = 0x03
        ENERGY      = 0x04
        CURR        = 0x05
        HUMIDITY    = 0x06
        VID         = 0x10
        INTRUSION   = 0x11
        BEEP_ENABLE = 0x18
        UNKNOWN     = 0xFFFFFFFF

        @property
        def units(self):
            """
            The units associated with this sensor reading type.
            """
            try:
                return {
                    self.IN.value : "V",
                    self.FAN.value : "RPM",
                    self.TEMP.value : "°C",
                    self.POWER.value : "W",
                    self.ENERGY.value : "J",
                    self.CURR.value : "A",
                    self.HUMIDITY.value : "%",
                }[self.value]
            except Exception as ex:
                return ""

        @property
        def type(self):
            """
            The name of the type of sensor reading.
            """
            try:
                return {
                    self.IN.value : "voltage",
                    self.FAN.value : "fanspeed",
                    self.TEMP.value : "temperature",
                    self.POWER.value : "power",
                    self.ENERGY.value : "energy",
                    self.CURR.value : "current",
                    self.HUMIDITY.value : "humidity",
                    self.VID.value : "vid",
                    self.INTRUSION.value : "intrusion",
                    self.BEEP_ENABLE.value : "beep"
                }[self.value]
            except Exception as ex:
                return "unknown"
